///
/// $Release: $
/// $Copyright: copyright(c) 2011 kuwata-lab.com all rights reserved $
/// $License: MIT License $
///

/*
 cmdopt.js - command option parser for node.js

 example1:
    // parser object
    cmdopt = require('./cmdopt.js');
    var parser = new cmdopt.Parser();
    // option definitions
    parser.option("-h", "--help").desc("show help message");
    parser.option("-f", "--file").arg("FILE").desc("file name");
    parser.option("-q").name("quiet").desc("quiet mode");
    parser.option("--verbose").desc("verbose mode");
    // parse command-line options
    var args = ["-hqf", "file.txt", "--verbose", "aaa", "bbb"];
    try {
      var opts = parser.parse(args);
      console.log(opts);  //=> { help: true, quiet: true, file: 'file.txt', verbose: true }
      console.log(args);  //=> [ 'aaa', 'bbb' ]
      // show help message
      if (opts.help) require('system').print(parser.helpMessage());
    }
    catch (ex) {
      if (! (ex instanceof cmdopt.ParseError)) throw ex;
      require('system').stderr.write(ex.message+"\n");
    }

 example2:
    // short and long options
    parser.option("-h", "--help").desc("show help message");
    // only short option, and add option name instead of long option
    parser.option("-v").name("version").desc("show version number");
    // option which takes argument
    parser.option("-f", "--file").arg("PATH").desc("file name")
    // option which can take optional argument
    parser.option("-W").arg("N", false).desc("warning (default N=2)");
    // argument data type
    parser.option("-W").arg("N").type("int");  // str/int/float/bool
    // argument pattern
    parser.option("-W").arg("N").pattern(/^\d+$/, "integer expected.");
    // custom validator of argument
    parser.option("-W").arg("N").validator(function(val) {
      // throw error message when validation failed
      if (! val.match(/^\d+$/) throw "integer expected.";
      n = parseInt(val);
      if (n < 1 || 3 < val) throw "should be between 1 to 3";
      // returned value (except undefined) will be used as argument value
      return n;
    })

 example3:
    // Regard long options as properties.
    // (User can specify any long options and they are regarded as properties.)
    var args = ["-q", "--help", "--foo=bar", "--baz=false"];
    var ret = parser.parseWithProperties(args);
    var opts = ret[0];
    var props = ret[1];
    console.log(opts); //=> { quiet: true }
    console.log(props); //=> { help: true, foo: 'bar', baz: false }

*/

"use_strict";


function classdef(constructor, prototype, imethods, cmethods) {
  function define(obj, funcs) {
    for (var i = 0, n = funcs.length; i < n; i++) {
      var func = funcs[i];
      if (func) obj[func.name] = func;
    }
  }
  if (prototype) constructor.prototype = prototype;
  if (cmethods) define(constructor, cmethods);
  if (imethods) define(constructor.prototype, imethods);
  return constructor;
}


function errordef(constructor, prototype, imethods, cmethods) {
  if (! prototype) prototype = new Error();
  prototype.name = constructor.name;
  return classdef(constructor, prototype, imethods, cmethods);
}


var ParseError = errordef(

  function ParseError(message) {
    Error.call(this, message);
    this.message = message;      // necessary. why?
  }

);


var DefinitionError = errordef(

  function DefinitionError(message) {
    Error.call(this, message);
    this.message = message;      // necessary. why?
  }

);


var Definition = classdef(

  function Definition() {
  },

  {
    name         : null,
    opt_short    : null,
    opt_long     : null,
    arg_name     : null,
    arg_required : false,
    pattern      : null,
    pattern_errmsg : null,
    type         : null,
    desc         : null,
    validator    : null
  }

);


var DefinitionBuilder = classdef(

  function DefinitionBuilder() {
    this.definition = new Definition();
  },

  {},

  [
    function option() {
      var optdef = this.definition;
      for (var i = 0, n = arguments.length; i < n; i++) {
        var arg = arguments[i];
        if (arg.match(/^-(\w)$/)) {
          optdef.opt_short = arg.slice(1);
        }
        else if (arg.match(/^--(\w[-\w]*)$/)) {
          optdef.opt_long = arg.slice(2);
        }
        else {
          throw new DefinitionError(arg + ": unknown option.");
        }
      }
      if (! optdef.name) optdef.name = optdef.opt_long || optdef.opt_short;
      return this;
    },
    
    function name(name) {
      this.definition.name = name;
      return this;
    },
    
    function arg(name, required) {
      this.definition.arg_name = name;
      this.definition.arg_required = required == null ? true : required;
      return this;
    },

    function pattern(regexp, errmsg) {
      this.definition.pattern = regexp;
      this.definition.pattern_errmsg = errmsg;
      return this;
    },
    
    function type(typename, errmsg) {
      var found = typename in Parser.prototype.TYPE_VALIDATORS;
      if (! found) throw new DefinitionError("type(): '" + typename + "': unknown type name.");
      this.definition.type = typename;
      if (errmsg != null) this.definition.type_errmsg = errmsg;
      return this;
    },
    
    function desc(msg) {
      this.definition.desc = msg;
      return this;
    },
    
    function validator(func) {
      this.definition.validator = func;
      return this;
    },
    
    null],

  []
);


var Parser = classdef(

  function Parser() {
    this.definitions = [];
  },

  {
    MAX_WIDTH: 30,
    MIN_WIDTH: 10,
    TYPE_VALIDATORS: {
      "str": function(val) {
        return val;
      },
      "int": function(val) {
        if (val.match(/^\d+$/)) return parseInt(val);
        throw "integer expected.";
      },
      "float": function(val) {
        if (val.match(/^\d+\.\d+$/)) return parseFloat(val);
        throw "float expected.";
      },
      "bool": function(val) {
        if (val.match(/^(true|yes|on)$/)) return true;
        if (val.match(/^(false|no|off)$/)) return false;
        throw "boolean expected.";
      }
    }
  },

  [
    function option() {
      var builder = new DefinitionBuilder();
      this.definitions.push(builder.definition);
      return builder.option.apply(builder, arguments);
    },

    function getDefinition(opt) {
      var optdefs = this.definitions;
      for (var i = 0, n = optdefs.length; i < n; i++) {
        var optdef = optdefs[i];
        if (optdef.opt_short === opt || optdef.opt_long === opt) {
          return optdef;
        }
      }
      return null;
    },

    function parse(args) {
      var options = {};
      this._parse(args, options);
      return options;
    },

    function parseWithProperties(args) {
      var options = {}, properties = {};
      this._parse(args, options, properties);
      return [options, properties];
    },

    function _parse(args, options, properties) {
      while (args[0] && args[0][0] === "-") {
        var arg = args.shift();
        if (arg === "--")        break;
        else if (arg[1] !== "-") this._parseShortOptions(arg, options, args);
        else if (! properties)   this._parseLongOption(arg, options, args);
        else                     this._parseProperty(arg, properties, args);
      }
    },

    function _parseShortOptions(opt, options, args) {
      for (var i = 1, optlen = opt.length; i < optlen; i++) {
        var ch = opt[i];
        var optdef = this.getDefinition(ch);
        if (! optdef) throw new ParseError("-"+ch+": unknown option.");
        var name = optdef.name;
        if (! optdef.arg_name) {
          options[name] = true;
          this._validate(optdef, options, name, "-" + ch); 
        }
        else if (optdef.arg_required) {
          if (i + 1 < optlen)   options[name] = opt.substring(i + 1);
          else if (args.length) options[name] = args.shift();
          else throw new ParseError("-"+ch+": argument required.");
          this._validate(optdef, options, name,
                         "-" + ch + " " + options[name]); 
          break;
        }
        else {
          options[name] = i + 1 < optlen ? opt.substring(i + 1) : true;
          this._validate(optdef, options, name,
                         "-" + ch + (i + 1 < optlen ? options[name] : ""));
          break;
        }
      }
    },

    function _parseLongOption(opt, options, args) {
      var m = opt.match(/^--(\w[-\w]*)(=.*)?$/);
      if (! m) throw new ParseError(opt + ": invalid option.");
      var key = m[1];
      var val = m[2] ? m[2].substring(1) : true;
      var optdef = this.getDefinition(key);
      if (! optdef) throw new ParseError("--"+key+": unknown option.");
      options[optdef.name] = val;
      this._validate(optdef, options, optdef.name, opt); 
    },

    function _parseProperty(opt, properties, args) {
      var m = opt.match(/^--(\w[-\w]*)(=.*)?$/);
      if (! m) throw new ParseError(opt + ": invalid property.");
      var key = m[1];
      var val = m[2] ? util.toValue(m[2].substring(1)) : true;
      properties[key] = val;
    },

    function _validate(optdef, options, name, cmdopt) {
      var val = options[name];
      if (optdef.pattern) {
        if (val !== true && ! val.match(optdef.pattern)) {
          var errmsg = optdef.pattern_errmsg || "invalid value.";
          throw new ParseError(cmdopt + ": " + errmsg);
        }
      }
      if (optdef.type) {
        if (val !== true) {
          var func = this.TYPE_VALIDATORS[optdef.type];
          if (! func) throw new DefinitionError("'" + optdef.type + "': unknown arg type.");
          try {
            options[name] = func(val);
          }
          catch (errmsg) {
            if (optdef.type_errmsg) errmsg = optdef.type_errmsg;
            throw new ParseError(cmdopt + ": " + errmsg);
          }
        }
      }
      if (optdef.validator) {
        try {
          val = optdef.validator(val);
          if (val !== undefined) options[name] = val;
        }
        catch (errmsg) {
          throw new ParseError(cmdopt + ": " + errmsg);
        }
      }
    },

    function helpMessage(width, indent, sep) {
      var pairs = [];
      var i, optdef;
      for (i = 0; optdef = this.definitions[i]; i++) {
        if (optdef.desc) {
          pairs.push([this._formatOption(optdef), optdef.desc]);
        }
      }
      if (width  == null) width  = this._calcWidth(pairs);
      if (indent == null) indent = "  ";
      if (sep    == null) sep    = " : ";
      return this._buildHelpMessage(pairs, width, indent, sep);
    },

    function _calcWidth(pairs) {
      var max = 0, pair;
      for (var i = 0; pair = pairs[i]; i++) {
        if (max < pair[0].length) max = pair[0].length;
      }
      return max > this.MAX_WIDTH ? this.MAX_WIDTH :
             max < this.MIN_WIDTH ? this.MIN_WIDTH :
             max;
    },

    function _buildHelpMessage(pairs, width, indent, sep) {
      var msg = "", pair;
      for (var i = 0; pair = pairs[i]; i++) {
        var s = pair[0], desc = pair[1];
        //s += s.length <= width ?
        //     util.repeat(" ", width - s.length) + sep :
        //     sep + "\n" + indent + util.repeat(" ", width);
        s += util.repeat(" ", width - s.length) + sep;
        msg += indent + s + desc + "\n";
      }
      return msg;
    },

    function _formatOption(optdef) {
      var opt_s = optdef.opt_short;
      var opt_l = optdef.opt_long;
      var arg   = optdef.arg_name;
      var arg_s = "", arg_l = "";
      if (arg) {
        arg_s = optdef.arg_required ? " " + arg : "[" + arg + "]";
        arg_l = optdef.arg_required ? "=" + arg : "[=" + arg + "]";
      }
      var s = "";
      if (opt_s && opt_l) s = "-" + opt_s + ", --" + opt_l + arg_l;
      else if (opt_s)     s = "-" + opt_s + arg_s;
      else if (opt_l)     s = "--" + opt_l + arg_l;
      else throw "unreachable";
      return s;
    },

    null]
);


var util = {

  repeat: function repeat(str, n) {
    var s = "";
    while (--n >= 0) s += str;
    return s;
  },

  toValue: function toValue(val) {
    if (val === "true")          return true;
    if (val === "false")         return false;
    if (val.match(/^\d+$/))      return parseInt(val);
    if (val.match(/^\d+\.\d+$/)) return parseFloat(val);
    if (val.match(/^".*"$/))     return val.slice(1, -1);
    if (val.match(/^'.*'$/))     return val.slice(1, -1);
    return val;
  }

};


var cmdopt = typeof(exports) === "object" ? exports : {};
cmdopt.Parser      = Parser;
cmdopt.ParseError  = ParseError;
cmdopt.DefinitionError = DefinitionError;
cmdopt._Definition = Definition;
cmdopt._DefinitionBuilder = DefinitionBuilder;
